mod cards;

use async_std::{prelude::*, sync::RwLock};
use cards::*;
use itertools::Itertools;
use rand::{seq::SliceRandom, thread_rng};
use serde::{Deserialize, Serialize};
use std::{collections::HashMap, sync::Arc};
use tide::{Body, Redirect, Request, Response};
use tide_websockets::{Message, WebSocket, WebSocketConnection};
use uuid::Uuid;

#[derive(Deserialize)]
#[serde(tag = "type")]
enum ClientMessage {
    Chat { message: String },
    CreateGame { name: String },
    JoinGame { name: String },
    StartGame,
    EndTurn,
    PlayCard { name: String, index: usize },
    GainCard { name: String, index: usize },
    BuyCard { index: usize },
    DrawCard,
    Discard { index: usize },
    TrashHand { index: usize },
}

#[derive(Serialize)]
#[serde(tag = "type")]
enum ServerMessage {
    Chat {
        sender: String,
        message: String,
    },
    PlayerJoined {
        player: String,
    },
    GameState {
        state: GameState,
        players: Vec<PlayerState>,
        active_player: usize,
        supply: Vec<PileState>,
        turn_state: TurnState,
        trash: Vec<Card>,
    },
    GameSetup {
        setup: GameSetup,
    },
    PlayerHand {
        hand: Vec<Card>,
    },
    PlayerId {
        id: usize,
    },
    Notification {
        text: String,
    },
    GameOver {
        score: Vec<(usize, u32)>,
    },
}

#[derive(Clone, Serialize)]
struct GameSetup {
    deck: Vec<Card>,
}

#[derive(Serialize)]
struct PlayerState {
    id: usize,
    name: String,
    draw_pile_count: usize,
    hand_count: usize,
    discard_pile: Option<String>,
    played_cards: Vec<String>,
}

#[derive(Serialize)]
struct PileState {
    name: Card,
    count: usize,
}

#[derive(Clone, Serialize)]
struct TurnState {
    actions: u32,
    buys: u32,
    coin: u32,
}

impl Default for TurnState {
    fn default() -> Self {
        TurnState {
            actions: 1,
            buys: 1,
            coin: 0,
        }
    }
}

struct Player {
    id: String,
    name: String,
    draw_pile: Vec<Card>,
    hand: Vec<Card>,
    discard_pile: Vec<Card>,
    played_cards: Vec<Card>,
}

impl Player {
    pub fn new(id: String, name: String) -> Self {
        Self {
            id,
            name,
            draw_pile: vec![],
            hand: vec![],
            discard_pile: vec![],
            played_cards: vec![],
        }
    }

    pub fn draw(&mut self, count: usize) {
        for _ in 0..count {
            if self.draw_pile.len() == 0 {
                self.draw_pile.append(&mut self.discard_pile);
                self.draw_pile.shuffle(&mut thread_rng());
            }

            if self.draw_pile.len() > 0 {
                self.hand.push(self.draw_pile.pop().unwrap());
            }
        }
    }

    pub fn discard(&mut self, index: usize) {
        self.discard_pile.push(self.hand.remove(index));
    }
}

#[derive(Copy, Clone, PartialEq, Serialize)]
enum GameState {
    Setup,
    InProgress,
    Over,
}

impl Default for GameSetup {
    fn default() -> GameSetup {
        GameSetup {
            deck: vec![
                copper(),
                copper(),
                copper(),
                copper(),
                copper(),
                copper(),
                copper(),
                estate(),
                estate(),
                estate(),
            ],
        }
    }
}

struct Game {
    players: Vec<Player>,
    state: GameState,
    setup: GameSetup,
    connections: HashMap<Uuid, (String, WebSocketConnection)>,
    active_player: usize,
    supply: Vec<(Card, usize)>,
    trash: Vec<Card>,
    turn_state: TurnState,
}

impl Game {
    fn new(player: Player) -> Self {
        Self {
            players: vec![player],
            state: GameState::Setup,
            setup: GameSetup::default(),
            connections: HashMap::new(),
            active_player: 0,
            supply: vec![],
            trash: vec![],
            turn_state: TurnState::default(),
        }
    }

    fn start(&mut self) {
        match self.state {
            GameState::Setup => {
                self.state = GameState::InProgress;

                for p in self.players.iter_mut() {
                    (*p).discard_pile = self.setup.deck.clone();
                    p.draw(5);
                }

                let victory_qty = match self.players.len() {
                    x if x <= 2 => 8,
                    _ => 12,
                };

                self.supply = vec![
                    (copper(), 60 - self.players.len() * 7),
                    (silver(), 40),
                    (gold(), 30),
                    (estate(), victory_qty),
                    (duchy(), victory_qty),
                    (province(), victory_qty),
                    (curse(), 10),
                    (
                        Card {
                            name: "Cellar".into(),
                            cost: 2,
                            types: vec![CardType::Action(|| {})],
                        },
                        10,
                    ),
                    (
                        Card {
                            name: "Moat".into(),
                            cost: 2,
                            types: vec![CardType::Action(|| {})],
                        },
                        10,
                    ),
                    (
                        Card {
                            name: "Village".into(),
                            cost: 3,
                            types: vec![CardType::Action(|| {})],
                        },
                        10,
                    ),
                    (
                        Card {
                            name: "Merchant".into(),
                            cost: 3,
                            types: vec![CardType::Action(|| {})],
                        },
                        10,
                    ),
                    (
                        Card {
                            name: "Workshop".into(),
                            cost: 3,
                            types: vec![CardType::Action(|| {})],
                        },
                        10,
                    ),
                    (
                        Card {
                            name: "Smithy".into(),
                            cost: 4,
                            types: vec![CardType::Action(|| {})],
                        },
                        10,
                    ),
                    (
                        Card {
                            name: "Remodel".into(),
                            cost: 4,
                            types: vec![CardType::Action(|| {})],
                        },
                        10,
                    ),
                    (
                        Card {
                            name: "Militia".into(),
                            cost: 4,
                            types: vec![CardType::Action(|| {})],
                        },
                        10,
                    ),
                    (
                        Card {
                            name: "Market".into(),
                            cost: 5,
                            types: vec![CardType::Action(|| {})],
                        },
                        10,
                    ),
                    (
                        Card {
                            name: "Mine".into(),
                            cost: 5,
                            types: vec![CardType::Action(|| {})],
                        },
                        10,
                    ),
                ];
            }

            _ => {} // Ignore if game is not in setup state
        }
    }

    fn end_turn(&mut self) {
        match self.players.get_mut(self.active_player) {
            None => {}
            Some(p) => {
                p.discard_pile.append(&mut p.played_cards);
                p.discard_pile.append(&mut p.hand);
                p.draw(5);
            }
        }

        self.end_game();

        self.active_player += 1;
        self.active_player %= self.players.len();
        self.turn_state = TurnState::default();
    }

    // Check if the end game condition is reached and finish the game if so,
    // sending a message to all clients.
    fn end_game(&mut self) {
        let provinces = self
            .supply
            .iter()
            .any(|(c, n)| (*c).name == "Province" && *n == 0);
        let supply = self.supply.iter().filter(|(_, n)| *n == 0).count() > 2;

        if supply || provinces {
            self.state = GameState::Over;

            for p in self.players.iter_mut() {
                p.draw_pile.append(&mut p.played_cards);
                p.draw_pile.append(&mut p.hand);
                p.draw_pile.append(&mut p.discard_pile);
            }
        }
    }

    fn is_active_player(&self, player_id: &str) -> bool {
        match self.players.get(self.active_player) {
            None => false,
            Some(p) => p.id == *player_id,
        }
    }

    pub fn trash_hand(&mut self, player_number: usize, index: usize) {
        self.trash
            .push(self.players[player_number].hand.remove(index));
    }

    pub fn play_card(&mut self, player_number: usize, index: usize) {
        let player = self.players.get_mut(player_number).unwrap();
        let card = player.hand.remove(index);

        if let Some(coin) = card.treasure() {
            self.turn_state.coin += coin;
        }

        if let Some(_) = card.action() {
            self.turn_state.actions -= 1;
        }

        player.played_cards.push(card);
    }

    pub fn buy_card(&mut self, player_number: usize, index: usize) -> bool /*-> Result<(), &'static str>*/
    {
        if player_number != self.active_player {
            return false;
        }

        if self.turn_state.buys < 1 {
            return false;
        }

        if let Some(card) = self.supply.get(index).as_deref() {
            if card.0.cost <= self.turn_state.coin {
                let card = card.0.clone();

                self.supply.get_mut(index).unwrap().1 = self.supply.get(index).unwrap().1 - 1;
                self.turn_state.coin -= card.cost;
                self.turn_state.buys -= 1;
                self.players[player_number].discard_pile.push(card);
                return true;
                //return Ok(());
            }
        }
        //Err("Not enough coin");
        false
    }
}

#[derive(Clone, Default)]
struct State {
    games: Arc<RwLock<HashMap<Uuid, Game>>>,
}

async fn notify_players(game: &Game, text: String) {
    let sm = ServerMessage::Notification { text };
    broadcast(game, &sm).await;
}

async fn broadcast(game: &Game, sm: &ServerMessage) {
    for (_, (_, con)) in game.connections.iter() {
        con.send_json(&sm).await.unwrap();
    }
}

async fn broadcast_state(game: &Game) {
    let sm = ServerMessage::GameState {
        state: game.state,
        players: game
            .players
            .iter()
            .enumerate()
            .map(|(i, p)| PlayerState {
                id: i,
                name: html_escape::encode_text(&p.name).into(),
                draw_pile_count: p.draw_pile.len(),
                hand_count: p.hand.len(),
                discard_pile: p.discard_pile.last().map(|c| c.name.clone()),
                played_cards: p.played_cards.iter().map(|c| c.name.clone()).collect(),
            })
            .collect(),
        active_player: game.active_player,
        supply: game
            .supply
            .iter()
            .map(|(card, count)| PileState {
                name: card.clone(),
                count: count.clone(),
            })
            .collect(),
        turn_state: game.turn_state.clone(),
        trash: game.trash.clone(),
    };

    broadcast(&game, &sm).await;

    for p in game.players.iter() {
        let sm = ServerMessage::PlayerHand {
            hand: p.hand.clone(),
        };

        send_msg(&game, &p, &sm).await;
    }

    if game.state == GameState::Over {
        let sm = ServerMessage::GameOver {
            score: game
                .players
                .iter()
                .enumerate()
                .map(|(i, p)| {
                    let score = p.draw_pile.iter().fold(0, |acc, card| {
                        if let Some(points) = card.victory() {
                            acc + points
                        } else if let Some(_) = card.curse() {
                            acc - 1
                        } else {
                            acc
                        }
                    });
                    (i, score)
                })
                .sorted_by(|a, b| Ord::cmp(&b.1, &a.1))
                .collect::<Vec<_>>(),
        };
        broadcast(&game, &sm).await;
    }
}

async fn send_msg(game: &Game, player: &Player, sm: &ServerMessage) {
    for (_, (player_id, con)) in game.connections.iter() {
        if *player_id == (*player).id {
            con.send_json(&sm).await.unwrap();
        }
    }
}

#[async_std::main]
async fn main() -> Result<(), std::io::Error> {
    tide::log::with_level(tide::log::LevelFilter::Debug);

    let mut app = tide::with_state(State::default());

    app.with(tide::sessions::SessionMiddleware::new(
        tide::sessions::MemoryStore::new(),
        "1234567890abcdef1234567890abcdef".as_bytes(),
    ));

    app.at("/")
        .get(|_| async { Ok(Body::from_file("static/index.html").await?) });

    app.at("/static").serve_dir("static/")?;

    app.at("/game").post(|mut req: Request<State>| async move {
        let msg: ClientMessage = req.body_form().await?;
        let session = req.session_mut();

        match msg {
            ClientMessage::CreateGame { name } => {
                session.insert("player_name", name.clone())?;
                let player_id: String = session.id().clone().into();

                let id = Uuid::new_v4();
                let mut games = req.state().games.write().await;
                let player = Player::new(player_id, name);

                games.insert(id, Game::new(player));

                let url = format!("/game/{}", id.to_simple().to_string());
                let res: tide::Result = Ok(Redirect::new(url).into());
                res
            }
            _ => Ok(Response::new(400)),
        }
    });

    app.at("/game/:id").get(|req: Request<State>| async move {
        let games = req.state().games.read().await;
        let id = req.param("id")?;
        let game_id = Uuid::parse_str(&id).unwrap();
        let session = req.session();

        match games.get(&game_id) {
            None => return Ok(Response::new(404)),
            Some(_) => (),
        };

        match session.get::<String>("player_name") {
            None => {
                let url = format!("/game/{}/join", id);
                return Ok(Redirect::new(url).into());
            }
            Some(_) => (),
        };

        let mut res = Response::new(200);
        res.set_body(Body::from_file("static/game.html").await?);
        Ok(res)
    });

    app.at("/game/:id/join")
        .get(|_| async { Ok(Body::from_file("static/join.html").await?) })
        .post(|mut req: Request<State>| async move {
            let msg: ClientMessage = req.body_form().await?;
            let id: String = { req.param("id").unwrap().into() };
            let session = req.session_mut();

            match msg {
                ClientMessage::JoinGame { name } => {
                    session.insert("player_name", name.clone())?;

                    let url = format!("/game/{}", id);
                    let res: tide::Result = Ok(Redirect::new(url).into());
                    res
                }
                _ => Ok(Response::new(500)),
            }
        });

    app.at("/game/:id/ws").get(WebSocket::new(
        |req: Request<State>, mut stream| async move {
            let id = req.param("id")?;
            let game_id = Uuid::parse_str(&id).unwrap();
            let mut games = req.state().games.write().await;
            let game = games.get_mut(&game_id).unwrap();
            let session = req.session();

            let player_id: String = session.id().clone().into();

            let client_id = Uuid::new_v4();
            game.connections
                .insert(client_id, (player_id.clone(), stream.clone()));

            let player_name: String = match session.get("player_name") {
                Some(p) => p,
                None => return Ok(()),
            };

            if game.state == GameState::Setup && !game.players.iter().any(|p| p.id == player_id) {
                game.players
                    .push(Player::new(player_id.clone(), player_name.clone()));
            }

            let player_number = game
                .players
                .iter()
                .enumerate()
                .filter(|(_, p)| p.id == player_id)
                .fold(0, |_, (i, _)| i);
            let msg = ServerMessage::PlayerId { id: player_number };
            stream.send_json(&msg).await?;

            //Ensure the write locks are freed, possibly better to move to a function
            //with implicit drop?
            drop(game);
            drop(games);

            let games = req.state().games.read().await;
            let game = games.get(&game_id).unwrap();

            broadcast_state(&game).await;

            let msg = ServerMessage::GameSetup {
                setup: game.setup.clone(),
            };
            stream.send_json(&msg).await?;

            let sm = ServerMessage::PlayerJoined {
                player: html_escape::encode_text(&player_name).into(),
            };

            broadcast(&game, &sm).await;

            drop(game);
            drop(games);

            while let Some(Ok(Message::Text(input))) = stream.next().await {
                let msg: ClientMessage = serde_json::from_str(&input)?;

                match msg {
                    ClientMessage::Chat { message } => {
                        let sender: String = session.get("player_name").unwrap();
                        let sm = ServerMessage::Chat {
                            sender: html_escape::encode_text(&sender).into(),
                            message: html_escape::encode_text(&message).into(),
                        };

                        let games = req.state().games.read().await;
                        let game = games.get(&game_id).unwrap();

                        broadcast(&game, &sm).await;
                    }

                    ClientMessage::StartGame => {
                        let mut games = req.state().games.write().await;
                        let game = games.get_mut(&game_id).unwrap();

                        game.start();

                        broadcast_state(&game).await;
                    }

                    ClientMessage::EndTurn => {
                        let mut games = req.state().games.write().await;
                        let game = games.get_mut(&game_id).unwrap();

                        if game.is_active_player(&player_id) {
                            game.end_turn();
                            broadcast_state(&game).await;
                        }
                    }

                    ClientMessage::CreateGame { .. } => {}

                    ClientMessage::JoinGame { .. } => {}

                    ClientMessage::PlayCard { name: _, index } => {
                        let mut games = req.state().games.write().await;
                        let game = games.get_mut(&game_id).unwrap();

                        let card_name = game.players[player_number].hand[index].name.clone();
                        game.play_card(player_number, index);

                        notify_players(
                            &game,
                            format!("{} spielt {}", game.players[player_number].name, card_name),
                        )
                        .await;
                        broadcast_state(&game).await;
                    }

                    ClientMessage::BuyCard { index } => {
                        let mut games = req.state().games.write().await;
                        let game = games.get_mut(&game_id).unwrap();

                        if game.buy_card(player_number, index) {
                            if let Some(card) =
                                game.players[player_number].discard_pile.last().as_deref()
                            {
                                notify_players(
                                    &game,
                                    format!(
                                        "{} kauft {}",
                                        game.players[player_number].name, card.name
                                    ),
                                )
                                .await;
                                broadcast_state(&game).await;
                            }
                        }
                    }

                    ClientMessage::GainCard { name: _, index } => {
                        let mut games = req.state().games.write().await;
                        let game = games.get_mut(&game_id).unwrap();

                        game.supply.get_mut(index).unwrap().1 =
                            game.supply.get(index).unwrap().1 - 1;
                        let card = game.supply[index].0.clone();

                        game.players[player_number].discard_pile.push(card.clone());

                        notify_players(
                            &game,
                            format!("{} nimmt {}", game.players[player_number].name, card.name),
                        )
                        .await;
                        broadcast_state(&game).await;
                    }

                    ClientMessage::DrawCard => {
                        let mut games = req.state().games.write().await;
                        let game = games.get_mut(&game_id).unwrap();

                        game.players[player_number].draw(1);

                        notify_players(
                            &game,
                            format!("{} zieht eine Karte", game.players[player_number].name),
                        )
                        .await;
                        broadcast_state(&game).await;
                    }

                    ClientMessage::Discard { index } => {
                        let mut games = req.state().games.write().await;
                        let game = games.get_mut(&game_id).unwrap();

                        let card_name = game.players[player_number].hand[index].clone();
                        game.players[player_number].discard(index);

                        notify_players(
                            &game,
                            format!("{} legt {} ab", game.players[player_number].name, card_name),
                        )
                        .await;
                        broadcast_state(&game).await;
                    }

                    ClientMessage::TrashHand { index } => {
                        let mut games = req.state().games.write().await;
                        let game = games.get_mut(&game_id).unwrap();

                        let card_name = game.players[player_number].hand[index].clone();
                        game.trash_hand(player_number, index);

                        notify_players(
                            &game,
                            format!(
                                "{} entsorgt {}",
                                game.players[player_number].name, card_name
                            ),
                        )
                        .await;
                        broadcast_state(&game).await;
                    }
                }
            }

            let mut games = req.state().games.write().await;
            let game = games.get_mut(&game_id).unwrap();

            game.connections.remove(&client_id);

            Ok(())
        },
    ));

    app.listen("0.0.0.0:5000").await?;
    Ok(())
}
